//
// KÖLTSÉGVETÉS2 - Költségvetés program, aktív MegtakaritasiSzamla
//                 és Folyoszamla osztállyal.
//
#include <cstdio>
#include <cstdlib>
#include <iostream>
using namespace std;

// a számlák maximális száma
const int MAXSZAMLAK = 10;

// Folyoszamla - ez írja le a folyószámlákat
class Folyoszamla
{
  public:
    Folyoszamla(int kezdetiSzSz = 0)
         : szamlaSzam(kezdetiSzSz), egyenleg(0.0) {}

    // lekérdező függvények
    int szlaSzam() { return szamlaSzam; }
    double szamlaEgyenleg() { return egyenleg; }

    // tranzakciós függvények
    double befizetes(double osszeg)
    {
        egyenleg += osszeg;
        return egyenleg;
    }
    double penzfelvetel(double osszeg);

    // megjelenítő függvény önmaga megjelenítéséhez a 'cout' kimeneten
    void megjelenites()
    {
        cout << szamlaSzam << " számla"
             << " = "   << egyenleg
             << endl;
    }

  protected:
    unsigned szamlaSzam;
    double  egyenleg;
};
//
// penzfelvetel - ez a tagfüggvény túl nagy a beágyazott
//              definiáláshoz
double Folyoszamla::penzfelvetel(double osszeg)
{
    if (egyenleg < osszeg)
    {
        cout << "Nincs elég fedezet: egyenleg " << egyenleg
             << ", ellenőrizze a következőt: "  << osszeg
             << endl;
        return false;
    }
    else
    {
        egyenleg -= osszeg;

        // ha az egyenleg nagyon lecsökken...
        if (egyenleg < 500.00)
        {
            // ...szolgáltatási díjat számítunk fel
            egyenleg -= 0.20;
        }
    }
    return true;
}

// MegtakaritasiSzamla - nyilvánvaló, hogy ez mi
class MegtakaritasiSzamla
{
  public:
    MegtakaritasiSzamla(int kezdetiSzSz = 0)
         : szamlaSzam(kezdetiSzSz),
         egyenleg(0.0), penzfelvetelekSzama(0) {}

    // lekérdező függvények
    int szlaSzam() { return szamlaSzam; }
    double szamlaEgyenleg() { return egyenleg; }

    // tranzakciós függvények
    double befizetes(double osszeg)
    {
        egyenleg += osszeg;
        return egyenleg;
    }
    bool penzfelvetel(double osszeg);

    // megjelenítő függvény - önmaga megjelenítése a 'cout' kimeneten
    void megjelenites()
    {
        cout << szamlaSzam             << " számla"
             << " = "                  << egyenleg
             << " (nincs penzfelvétel = " << penzfelvetelekSzama
             << ")" << endl;
    }
  protected:
    unsigned szamlaSzam;
    double  egyenleg;
    int   penzfelvetelekSzama;
};
bool MegtakaritasiSzamla::penzfelvetel(double osszeg)
{
    if (egyenleg < osszeg)
    {
        cout << "Nincs elég fedezet: az egyenleg " << egyenleg
             << ", penzfelvétel: "        << osszeg
             << endl;
        return false;
    }
    else
    {
        // havonta egynél több penzfelvétel után...
        if (++penzfelvetelekSzama > 1)
        {
            // ...$5-os díjat számítunk fel
            egyenleg -= 5.00;
        }

        // most végrehajtjuk a penzfelvételt
        egyenleg -= osszeg;
    }
    return true;
}

// prototípusdeklarációk
void feldolgoz(Folyoszamla* mFolyoszamla);
void feldolgoz(MegtakaritasiSzamla*  pMegtakaritasiSzamla);

// a folyó- és megtakarítási számlák objektumai
Folyoszamla* folyoSzamlak[MAXSZAMLAK];
MegtakaritasiSzamla*  megtakSzamlak[MAXSZAMLAK];

// main - összegyűjti a kezdeti be- és kimeneti összegeket
int main(int nArgSzama, char* mszArg[])
{
    // a ciklus 'X' vagy 'x' megadásáig fut
    int folyoSzamlakSzama = 0;  // megszámoljuk a számlákat
    int megtakSzamlakSzama = 0;
    char   szamlaTipus;   // M vagy F
    while(true)
    {
        cout << "M - Megtakarítasi számla, "
             << "F - Folyószamla, "
             << "X - kilépés:";
        cin >> szamlaTipus;

        // ha a felhasználó X-et ad meg, kilépünk a ciklusból
        if (szamlaTipus == 'x' || szamlaTipus == 'X')
        {
            break;
        }

        // különben a típusának megfelelően kezeljük a számlát
        switch (szamlaTipus)
        {
          // folyószámla
          case 'f':
          case 'F':
            if (folyoSzamlakSzama < MAXSZAMLAK)
            {
                int szla;
                cout << "Add meg a számlaszámot:";
                cin  >> szla;
                folyoSzamlak[folyoSzamlakSzama] =
                                     new Folyoszamla(szla);
                feldolgoz(folyoSzamlak[folyoSzamlakSzama]);
                folyoSzamlakSzama++;
            }
            else
            {
                cout << "Nincs már hely folyószámlák számára"
                     << endl;
            }
            break;

          // megtakarítási számla
          case 'm':
          case 'M':
            if (megtakSzamlakSzama < MAXSZAMLAK)
            {
                int szla;
                cout << "Add meg a számlaszámot:";
                cin  >> szla;
                megtakSzamlak[megtakSzamlakSzama] = new MegtakaritasiSzamla(szla);
                feldolgoz(megtakSzamlak[megtakSzamlakSzama]);
                megtakSzamlakSzama++;
            }
            else
            {
                cout << "Nincs már hely megtakarítási számlák számára"
                     << endl;
            }
            break;

          default:
            cout << "Nem értem." << endl;
      }
    }

    // most kiírjuk az összegeket
    double folyoszlaOsszeg = 0;    // az összes folyószámla összege
    cout << "Folyoszámlák:\n";
    for (int i = 0; i < folyoSzamlakSzama; i++)
    {
        folyoSzamlak[i]->megjelenites();
        folyoszlaOsszeg += folyoSzamlak[i]->szamlaEgyenleg();
    }

    double megtakszlaOsszeg = 0;    // az összes megtakarítási számla összege
    cout << "Megtakarítási számlák:\n";
    for (int j = 0; j < megtakSzamlakSzama; j++)
    {
        megtakSzamlak[j]->megjelenites();
        megtakszlaOsszeg += megtakSzamlak[j]->szamlaEgyenleg();
    }

    double osszeg = folyoszlaOsszeg + megtakszlaOsszeg;
    cout << "Folyószámlák összege = "
         << folyoszlaOsszeg
         << endl;

    cout << "Megtakarítási számlák összege = "
         << megtakszlaOsszeg
         << endl;

    cout << "Teljes összeg         = "
         << osszeg
         << endl;

    // várakozás a program leállításával, amíg a felhasználó készen áll,
    // hogy a felhasználó láthassa a program eredményeit
    cout << "A folytatáshoz nyomd meg az Entert..." << endl;
    cin.ignore(10, '\n');
    cin.get();
    return 0;
}

// feldolgoz(Folyoszamla) - egy folyószámla adatainak megadása
void feldolgoz(Folyoszamla* mFolyoszamla)
{
    cout << "Befizetéshez pozitív számot,\n"
         << "pénzfelvételhez negatív számot, befejezéshez 0-t adj meg"
         << endl;
    double tranzakcio;
    while(true)
    {
        cout << ":";
        cin >> tranzakcio;
        if (tranzakcio == 0)
        {
            break;
        }

        // befizetés
        if (tranzakcio > 0)
        {
          mFolyoszamla->befizetes(tranzakcio);
        }

        // penzfelvétel
        if (tranzakcio < 0)
        {
          mFolyoszamla->penzfelvetel(-tranzakcio);
        }
    }
}

// feldolgoz(MegtakaritasiSzamla) - egy megtakarítási számla adatainak megadása
void feldolgoz(MegtakaritasiSzamla* pMegtakaritasiSzamla)
{
    cout << "Befizetéshez pozitív számot,\n"
         << "pénzfelvételhez negatív számot, befejezéshez 0-t adj meg"
         << endl;
    double tranzakcio;
    while(true)
    {
        cout << ":";
        cin >> tranzakcio;
        if (tranzakcio == 0)
        {
            break;
        }

        // befizetés
        if (tranzakcio > 0)
        {
          pMegtakaritasiSzamla->befizetes(tranzakcio);
        }

        // penzfelvétel
        if (tranzakcio < 0)
        {
          pMegtakaritasiSzamla->penzfelvetel(-tranzakcio);
        }
    }
}
