// KÖLTSÉGVETÉS3 - Költségvetés program öröklődéssel
//           és késői kötéssel (polimorfizmussal). Egyetlen
//           függvénnyel kezelhetők a folyó- és a
//           megtakarítási számlák is (és bármilyen más
//           számla, amit később esetleg kitalálsz).
//
//           Emellett ez a verzió egy láncolt listában
//           tárolja a számlákat rögzített tömb helyett,
//           hogy ne legyen rögzített maximális számra
//           korlátozva az objektumok száma.

#include <cstdio>
#include <cstdlib>
#include <iostream>

#include "SzamlakLancoltListaja.h"
using namespace std;
using namespace Listak;

// Szamla - ez az absztrakt osztály magába foglalja a
//          két számlatípus, a Folyószámla és a Megtakarítási
//          számla közös tulajdonságait. Nincs viszont benne a
//          penzfelvetel(), mert az a kettőben eltérő
class Szamla
{
  public:
    Szamla(SzamlakLancoltListaja* mLista, int szlaSzam)
       : csomopont(mLista, this), szamlaSzam(szlaSzam), egyenleg(0)
    {
      darabszam++;
    }

    // lekérdező függvények
    int szlaSzam() { return szamlaSzam; }
    double szlaEgyenleg() { return egyenleg; }
    static int szamlakSzama() { return darabszam; }

    // tranzakciós függvények
    void befizetes(double osszeg) { egyenleg += osszeg; }
    virtual bool penzfelvetel(double osszeg)
    {
        if (egyenleg < osszeg )
        {
             cout << "Nincs elég fedezet: az egyenleg " << egyenleg
                  << ", penzfelvétel: "                << osszeg
                  << endl;
              return false;
        }
        egyenleg -= osszeg;
        return true;
    }

    // megjelenítő függvény önmaga megjelenítéséhez a 'cout' kimeneten
    void megjelenites()
    {
        cout << tipus()
             << szamlaSzam << " számla" 
             << " = "   << egyenleg
             << endl;
    }
    virtual const char* tipus() = 0;

  protected:
    // a csomópont tartalmazza a láncolt lista adatait
    Csomopont csomopont;

    static int darabszam;             // a számlák száma
    unsigned szamlaSzam;
    double   egyenleg;
};

// tárhely lefoglalása a statikus tagok számára
int Szamla::darabszam = 0;

// Folyoszamla - ez az osztály a folyószámlák tulajdonságait
//            tartalmazza. Hát nem sok minden maradt, ugye?
class Folyoszamla : public Szamla
{
  public:
    Folyoszamla(SzamlakLancoltListaja* mLL, unsigned szlaSzam)
      : Szamla(mLL, szlaSzam) { }

    // túlterheljük a tisztán virtuális függvényeket
    virtual bool penzfelvetel(double osszeg);
    virtual const char* tipus() { return "Folyoszámla"; }
};

// penzfelvétel - felüldefiniáljuk a Szamla::penzfelvetel() tag-
//                függvényt, hogy felszámoljon 20 centet pénzfelvételenként,
//                ha az egyenleg $500 alatt van
bool Folyoszamla::penzfelvetel(double osszeg)
{
    bool siker = Szamla::penzfelvetel(osszeg);

    // ha az egyenleg nagyon lecsökken, szolgáltatási díjat számítunk fel
    if (siker && egyenleg < 500.00)
    {
        egyenleg -= 0.20;
    }
    return siker;
}

// MegtakaritasiSzamla - ugyanaz, mint a Folyoszamla esetében, de ennek
// is van egy egyedi adattagja
class MegtakaritasiSzamla : public Szamla
{
  public:

    MegtakaritasiSzamla(SzamlakLancoltListaja* mLL, unsigned szlaSzam)
      : Szamla(mLL, szlaSzam), penzfelvetelekSzama(0) {}

    // tranzakciós függvények
    virtual bool penzfelvetel(double osszeg);
    virtual const char* tipus() { return "MegtakaritasiSzamla"; }

  protected:
    int penzfelvetelekSzama;
};

// penzfelvétel - felüldefiniáljuk a Szamla::penzfelvetel() tag-
//                függvényt, hogy $5.00 díjat számítson fel
//                a hónap első pénzfelvétele után
bool MegtakaritasiSzamla::penzfelvetel(double osszeg)
{
    if (++penzfelvetelekSzama > 1)
    {
        egyenleg -= 5.00;
    }
    return Szamla::penzfelvetel(osszeg);
}

// prototípusdeklarációk
unsigned szlaSzamLekerese();
void     feldolgoz(Szamla* mSzamla);
void     szamlakBeolvasasa(SzamlakLancoltListaja* mLancoltLista);
void     eredmenyekMegjelenitese(SzamlakLancoltListaja* mLancoltLista);


// main - összegyűjti a kezdeti be- és kimeneti összegeket
int main(int nArgSzama, char* mszArg[])
{
    // létrehozunk egy láncolt listát, amelyhez hozzáadjuk a számlákat
    SzamlakLancoltListaja* mLancoltLista = new SzamlakLancoltListaja();

    // beolvassuk a számlákat a felhasználótól
    szamlakBeolvasasa(mLancoltLista);

    // megjelenítjük a számlák láncolt listáját
    eredmenyekMegjelenitese(mLancoltLista);

    // várakozás a program leállításával, amíg a felhasználó készen áll,
    // hogy a felhasználó láthassa a program eredményeit
    cout << "A folytatáshoz nyomd meg az Entert..." << endl;
    cin.ignore(10, '\n');
    cin.get();
    return 0;
}

// szamlakBeolvasasa - feltöltjük a számlák megadott tömbjét
void szamlakBeolvasasa(SzamlakLancoltListaja* mLancoltLista)
{
    Szamla* mSz;

    // a ciklus 'X' vagy 'x' megadásáig fut
    char   szamlaTipus;     // M vagy F
    while (true)
    {
        cout << "M - Megtakarítasi számla, "
             << "F - Folyószamla, X - kilépés:";
        cin >> szamlaTipus;
        switch (szamlaTipus)
        {
          case 'f':
          case 'F':
            mSz = new Folyoszamla(mLancoltLista, szlaSzamLekerese());
            break;

          case 'm':
          case 'M':
            mSz = new MegtakaritasiSzamla(mLancoltLista, szlaSzamLekerese());
            break;

          case 'x':
          case 'X':
            return;

          default:
            cout << "Nem értem.\n";
        }

        // most feldolgozzuk a létrehozott objektumot
        feldolgoz(mSz);
    }
}

// eredmenyekMegjelenitese - a számlák láncolt listájában talált számlákat
//                  jeleníti meg
void eredmenyekMegjelenitese(SzamlakLancoltListaja* mLancoltLista)
{
    // most kiírjuk az összegeket
    double osszeg = 0.0;
    cout << "\nA számlák összegei:\n";
    for (Csomopont* mC = Csomopont::elsoCsomopont(mLancoltLista);
               mC != 0;
               mC = mC->kovetkezoCsomopont())
    {
        Szamla* mSz = mC->aktualisSzamla();
        mSz->megjelenites();
        osszeg += mSz->szlaEgyenleg();
    }
    cout << "Teljes összeg = " << osszeg << "\n";
}

// szlaSzamLekerese - visszaadja a számlaszámot a számla létrehozásához
unsigned szlaSzamLekerese()
{
    unsigned szlaSzam;
    cout << "Add meg a számlaszámot:";
    cin  >> szlaSzam;
    return szlaSzam;
}

// feldolgoz(Szamla) - egy számla adatainak megadása
void feldolgoz(Szamla* mSzamla)
{
    cout << "Befizetéshez pozitív számot,\n"
         << "penzfelvételhez negatív számot, a befejezéshez 0-t adj meg\n";
    double tranzakcio;
    while(true)
    {
        cout << ":";
        cin >> tranzakcio;
        if (tranzakcio == 0)
        {
            break;
        }

        // befizetes
        if (tranzakcio > 0)
        {
            mSzamla->befizetes(tranzakcio);
        }
        // penzfelvetel
        if (tranzakcio < 0)
        {
            mSzamla->penzfelvetel(-tranzakcio);
        }
    }
}

