// KÖLTSÉGVETÉS5 - Megegyezik a többi Költségvetés programmal, azt leszámítva,
//                 hogy az STL listakonténert használja a költségvetési
//                 objektumok tárolásához (nem pedig egy rögzíttt tömböt
//                 vagy egy házi készítésű láncolt listát)

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <list>
using namespace std;

/** \class Szamla
 *  \brief egy absztrakt bankszámla.
 *  \details ez az absztrakt osztály azokat a tulajdonságokat
 *           tartalmazza, amelyek mindkét számlatípusra jellemzők:
 *           a Folyoszamla és a MegtakaritasiSzamla osztályra is.
 *           Hiányzik viszont belőle a penzfelvetel(), ami a kettőnél eltér
 */
class Szamla
{
  public:
    /** \brief Szamla konstruktor
     *
     * \param szlaSzam - a használandó számlaszám
     */
    Szamla(unsigned szlaSzam)
    {
      // az objektum adattagjainak inicializálása
      szamlaSzam = szlaSzam;
      egyenleg = 0;
      darabszam++;
    }

    // lekérdező függvények
    /** \brief szlaSzam - a számlaszámot adja vissza
     * \return int
     */
    int szlaSzam() { return szamlaSzam; }
    /** \brief szlaEgyenleg - a számlaegyenleget adja vissza
     * \return double
     */
    double szlaEgyenleg() { return egyenleg; }
    /** \brief szamlakSzama - a számlák számát adja vissza
     * \return int
     */
    static int szamlakSzama() { return darabszam; }

    // tranzakciós függvények
    /** \brief befizetes - egy összeg befizetése a számlára
     */
    void befizetes(double osszeg) { egyenleg += osszeg; }

    /** \brief penzfelvetel - egy összeg felvétele a
     *                        számláról
     *
     * A függvény először ellenőrzi, hogy van-e elég
     * rendelkezésre álló fedezet. Ha nincs, akkor
     * a függvény egy hibaüzenetet ír ki, és
     * false értéket ad vissza.
     * \return true értéked ad vissza a függvény megfelelő lefutása esetén
     */
    virtual bool penzfelvetel(double osszeg)
    {
        if (egyenleg < osszeg )
        {
             cout << "Nincs elég fedezet: az egyenleg " << egyenleg
                  << ", penzfelvétel: "                << osszeg
                  << endl;
              return false;
        }
        egyenleg -= osszeg;
        return true;
    }

    // megjelenítő függvény önmaga megjelenítéséhez a 'cout' kimeneten
    /** \brief megjelenites - az aktuális objektum megjelenítése
     *         a standard kimeneten
     */
    void megjelenites()
    {
        cout << tipus()
             << szamlaSzam << " számla" 
             << " = "   << egyenleg
             << endl;
    }
    /** \brief tipus - annak visszaadása, hogy milyen típusú
     *         Szamla objektum az aktuális objektum
     *
     * \return string, amely az osztály nevét tartalmazza
     *
     */
     virtual const char* tipus() { return "Szamla"; }

  protected:
    static int darabszam;/**< a számlaobjektumok száma*/
    unsigned szamlaSzam;/**< az aktuális számla*/
                           /**< száma */
    double   egyenleg;/**< az aktuális számla egyenlege*/
};

// tárhely lefoglalása a statikus tagok számára
int Szamla::darabszam = 0;

/** \class Folyoszamla
 * \brief a folyószámlák egyedi tulajdonságai.
 * \details A legtöbb tulajdonság a Szamla alaposztályban
 * található. A Szamla két alosztálya közül ez az egyik.
 * A Folyoszamla egy folyószámla tulajdonságait
 * valósítja meg mitikus bankunkban.
 */
class Folyoszamla : public Szamla
{
  public:
    Folyoszamla(unsigned szlaSzam) :
      Szamla(szlaSzam)
    { }

    // túlterheljük a tisztán virtuális függvényeket
    virtual bool penzfelvetel(double osszeg);
    virtual const char* tipus() { return "Folyoszámla"; }
};

/** \brief penzfelvetel - penzfelvétel folyószámláról
 * \details A folyószámla abban tér el egy standard
 * számlától, hogy felszámít 20 centet pénzfelvételenként,
 * ha az egyenleg $500 alatt van.
 *
 * \param osszeg - a felvenni kívánt összeg
 * \return bool true, ha sikerül a pénzfelvétel
 */
bool Folyoszamla::penzfelvetel(double osszeg)
{
    bool siker = Szamla::penzfelvetel(osszeg);

    // ha az egyenleg nagyon lecsökken, szolgáltatási díjat számítunk fel
    if (siker && egyenleg < 500.00)
    {
        egyenleg -= 0.20;
    }
    return siker;
}

/** \class MegtakaritasiSzamla
 * \brief egy megtakarítási számla
 *
 * \details A Szamla két alosztályának egyike.
 * A MegtakaritasiSzamla egy megtakarítási számla részleteit
 * valósítja meg mitikus bankunkban.
 */
class MegtakaritasiSzamla : public Szamla
{
  public:
    /** \brief konstruktor
     *
     * \param szlaSzam - számlaszám
     */
    MegtakaritasiSzamla(unsigned szlaSzam) : Szamla(szlaSzam)
    { penzfelvetelekSzama = 0; }

    // tranzakciós függvények
    virtual bool penzfelvetel(double osszeg);
    virtual const char* tipus() { return "MegtakaritasiSzamla"; }

  protected:
    int penzfelvetelekSzama;   /**< a pénzfelvételek száma */
};

/** \brief felüldefiniáljuk a Szamla::penzfelvetel() tag-
 *  \details függvényt, hogy $5.00 díjat számítson fel
 *           a hónap első pénzfelvétele után
 *
 * \param osszeg double - a felvenni kívánt összeg
 * \return bool true, ha sikerül a pénzfelvétel
 */
bool MegtakaritasiSzamla::penzfelvetel(double osszeg)
{
    if (++penzfelvetelekSzama > 1)
    {
        egyenleg -= 5.00;
    }
    return Szamla::penzfelvetel(osszeg);
}

/**
 * \brief egy számla mutatója
 */
typedef Szamla* SzamlaMut;

// prototipus declarations
unsigned szlaSzamLekerese();
void     feldolgoz(SzamlaMut mSzamla);
void     szamlakBeolvasasa(list<SzamlaMut>& szlaLista);
void     eredmenyekMegjelenitese(list<SzamlaMut>& szlaLista);

int main(int nArgSzama, char* mszArg[])
{
    // létrehozunk egy láncolt listát, amelyhez hozzáadjuk a számlákat
    list<SzamlaMut> szamlakListaja;

    // beolvassuk a számlákat a felhasználótól
    szamlakBeolvasasa(szamlakListaja);

    // megjelenítjük a számlák láncolt listáját
    eredmenyekMegjelenitese(szamlakListaja);

    // várakozás a program leállításával, amíg a felhasználó készen áll,
    // hogy a felhasználó láthassa a program eredményeit
    cout << "A folytatáshoz nyomd meg az Entert..." << endl;
    cin.ignore(10, '\n');
    cin.get();
    return 0;
}

/** \brief szamlakBeolvasasa - számlák megadása a billentyűzeten
 * \details Ez a függvény beolvassa a bemenetet a billentyűzetről.
 * Minden megadott M vagy F betűnél létrehoz egy új
 * MegtakaritasiSzamla vagy Folyoszamla számlaobjektumot, és hozzáadja azt
 * a számlák listájához. A bevitel X betűvel állítható le. Bármilyen más
 * bemenetről feltételezi, hogy befizetés (0-nál nagyobb
 * számoknál) vagy pénzfelvétel (0-nál kisebb számoknál).
 *
 * \param szlaLista list<SzamlaMut>& a számlaobjektumok
 *                listája, amelyet a szamlakBeolvasasa() létrehoz
 * \return void
 */
void szamlakBeolvasasa(list<SzamlaMut>& szlaLista)
{
    SzamlaMut mSz;

    // a ciklus 'X' vagy 'x' megadásáig fut
    char   szamlaTipus;     // M vagy F
    while (true)
    {
        cout << "M - Megtakarítasi számla, "
             << "F - Folyószamla, X - kilépés:";
        cin >> szamlaTipus;
        mSz = nullptr;
        switch (szamlaTipus)
        {
          case 'f':
          case 'F':
            mSz = new Folyoszamla(szlaSzamLekerese());
            break;

          case 'm':
          case 'M':
            mSz = new MegtakaritasiSzamla(szlaSzamLekerese());
            break;

          case 'x':
          case 'X':
            return;

          default:
            cout << "Nem értem.\n";
        }

        // most feldolgozzuk a létrehozott objektumot
        if (mSz)
        {
            szlaLista.push_back(mSz);
            feldolgoz(mSz);
        }
    }
}

/** \brief eredmenyekMegjelenitese - a számlák láncolt listájában
 *                                   talált számlákat jeleníti meg
 *
 * \param szlaLista list<SzamlaMut>& a számlák listája
 * \return void
 */
void eredmenyekMegjelenitese(list<SzamlaMut>& szlaLista)
{
    // most kiírjuk az összegeket
    double osszeg = 0.0;
    cout << "\nA számlák összegei:\n";

    // létrehozunk egy iterátort és bejárjuk a listát
    // (az iter típusa list<SzamlaMut>::iterator)
    for (list<SzamlaMut>::iterator iter = szlaLista.begin();
         iter != szlaLista.end();
         iter++)
    {
        SzamlaMut mSzamla = *iter;
        mSzamla->megjelenites();
        osszeg += mSzamla->szlaEgyenleg();
    }
    cout << "Teljes összeg = " << osszeg << "\n";
}

//
/** \brief szlaSzamLekerese - visszaadja a számlaszámot a
 *             számla létrehozásához
 * \return unsigned a felhasználó által megadott számlaszám
 */
unsigned szlaSzamLekerese()
{
    unsigned szlaSzam;
    cout << "Add meg a számlaszámot:";
    cin  >> szlaSzam;
    return szlaSzam;
}

/** \brief feldolgoz(Szamla) - egy számla adatainak megadása
 *
 * \param mSzamla SzamlaMut a feldolgozandó számla
 * \return void
 */
void feldolgoz(SzamlaMut mSzamla)
{
    cout << "Befizetéshez pozitív számot,\n"
         << "penzfelvételhez negatív számot, a befejezéshez 0-t adj meg\n";
    double tranzakcio;
    while(true)
    {
        cout << ":";
        cin >> tranzakcio;
        if (tranzakcio == 0)
        {
            break;
        }

        // befizetes
        if (tranzakcio > 0)
        {
            mSzamla->befizetes(tranzakcio);
        }
         // penzfelvetel
        if (tranzakcio < 0)
        {
            mSzamla->penzfelvetel(-tranzakcio);
        }
    }
}
